class BinaryIterator(object):
    def __init__(self, content):
        self.content = content

    @classmethod
    def open_file(self, path):
        return open(path, 'rb')

    @classmethod
    def decode(self, content):
        return content

    def __iter__(self):
        yield self.content
        raise StopIteration



import os


class DataFile(object):
    def __iter__(self):
        return self

    def __next__(self):
        raise NotImplementedError

    def _is_gzip(self):
        return os.path.splitext(self.filename)[-1] == '.gz'



import os

from akagi.log import logger


class DataSource(object):
    '''DataSource is an base class of all data sources
    '''

    def __init__(self, bundle):
        self.bundle = bundle

    def save(self, tar_dir, force=False):
        paths = []

        for d in self.bundle.data_files:
            path = os.path.expanduser(os.path.join(tar_dir, d.key))
            dirname = os.path.dirname(path)

            if not os.path.isdir(dirname):
                os.makedirs(dirname)

            if os.path.isfile(path):
                logger.debug("Skipped %(key)s" % ({"key": d.key}))
            else:
                with open(path, 'wb') as f:
                    f.write(d.raw_content)
                    logger.debug("Saved %(key)s to %(path)s" % ({"key": d.key, "path": path}))

            paths.append(path)

        return paths

    def __enter__(self):
        return self

    def __exit__(self, *exc):
        raise NotImplementedError

    def __iter__(self):
        return iter(self.bundle)



import os

import psycopg2

from akagi.data_source import DataSource
from akagi.data_file_bundles import S3DataFileBundle
from akagi.query import UnloadQuery
from akagi.log import logger


class RedshiftDataSource(DataSource):
    '''RedshiftDataSource replesents a set of row data as a result of query param.
    It uses UNLOAD command and intermediate Amazon S3 bucket.
    '''

    @classmethod
    def for_query(cls, query, schema, table, bucket_name, db_conf={}, sort=False, activate=True):
        bundle = S3DataFileBundle.for_table(bucket_name, schema, table)

        query = UnloadQuery.wrap(query, bundle, sort)

        return RedshiftDataSource(bundle, query, db_conf, activate)

    def __init__(self, bundle, query, db_conf={}, activate=True):
        super(RedshiftDataSource, self).__init__(bundle)
        self.query = query
        self.__db_conf = db_conf
        self.__pgpass = None

        if activate:
            self.activate()

    def activate(self):
        logger.info("Deleting old files on s3...")
        self.bundle.clear()
        logger.info("Sending query to Redshift")
        logger.debug("\n" + self.query.body + "\n")  # avoid logging unload query since it has raw credentials inside
        self._cursor.execute(self.query.sql)

    @property
    def _connection(self):
        return psycopg2.connect(**self._db_conf)

    @property
    def _cursor(self):
        return self._connection.cursor()

    @property
    def _db_conf(self):
        return {
                'host':     self.__db_conf.get('host') or os.getenv('REDSHIFT_DB_HOST', self._pgpass['db_host']),
                'user':     self.__db_conf.get('user') or os.getenv('REDSHIFT_DB_USER', self._pgpass['db_user']),
                'dbname':   self.__db_conf.get('dbname') or os.getenv('REDSHIFT_DB_NAME', self._pgpass['db_name']),
                'password': self.__db_conf.get('password') or os.getenv('REDSHIFT_DB_PASS', self._pgpass['db_pass']),
                'port':     self.__db_conf.get('port') or os.getenv('REDSHIFT_DB_PORT', self._pgpass['db_port'])
                }

    @property
    def _pgpass(self):
        if self.__pgpass is None:
            self.__pgpass = {}

            with open(os.path.expanduser(os.path.join('~', '.pgpass'))) as f:
                args = [s.strip() for s in f.read().split(':')]

                if len(args) == 5:
                    (
                            self.__pgpass['db_host'], self.__pgpass['db_port'],
                            self.__pgpass['db_name'], self.__pgpass['db_user'],
                            self.__pgpass['db_pass']) = args

        return self.__pgpass

    def __exit__(self, *exc):
        self.bundle.clear()
        return False



from akagi.iterators import CSVIterator, BinaryIterator


class FileFormat(object):
    CSV = 1
    BINARY = 2


class Iterator(object):
    @classmethod
    def open_file(self, path):
        raise NotImplementedError

    @classmethod
    def decode(self, content):
        raise NotImplementedError

    @classmethod
    def get_iterator_class(cls, file_format):
        if file_format in [FileFormat.CSV, 'csv']:
            return CSVIterator
        elif file_format in [FileFormat.BINARY, 'binary']:
            return BinaryIterator
        else:
            raise Exception("Unsupported file format %(file_format)s." % locals())



import os

from akagi.utils import gzip_decompress

from akagi.data_file import DataFile


class LocalDataFile(DataFile):
    def __init__(self, path, iterator_class):
        self.path = path
        self.iterator_class = iterator_class

    @property
    def filename(self):
        return os.path.basename(self.path)

    @property
    def raw_content(self):
        return open(self.path).read()

    @property
    def content(self):
        if self._is_gzip():
            content = gzip_decompress(open(self.path, 'rb').read())
            return self.iterator_class.decode(content)
        else:
            return self.iterator_class.open_file(self.path).read()

    def __iter__(self):
        return iter(self.iterator_class(self.content))



from akagi.data_source import DataSource
from akagi.data_file_bundles import S3DataFileBundle
from akagi.iterator import FileFormat


class S3DataSource(DataSource):
    '''S3DataSource replesents a set of files on Amazon S3 bucket.
    '''

    @classmethod
    def for_prefix(cls, bucket_name, prefix, file_format=FileFormat.BINARY):
        bundle = S3DataFileBundle(
                bucket_name,
                prefix=prefix,
                file_format=file_format
                )

        return S3DataSource(bundle)

    @classmethod
    def for_keys(cls, bucket_name, keys, file_format=FileFormat.BINARY):
        bundle = S3DataFileBundle(
                bucket_name,
                keys=keys,
                file_format=file_format
                )

        return S3DataSource(bundle)

    @classmethod
    def for_key(cls, bucket_name, key, file_format=FileFormat.BINARY):
        return S3DataSource.for_prefix(bucket_name, key, file_format)

    def __init__(self, bundle):
        self.bundle = bundle

    def __exit__(self, *exc):
        return False



import os
import glob

from akagi.iterator import Iterator, FileFormat
from akagi.data_file_bundle import DataFileBundle
from akagi.data_files import LocalDataFile


class LocalDataFileBundle(DataFileBundle):
    def __init__(self, directory_path, file_format=FileFormat.CSV):
        self.directory_path = os.path.expanduser(directory_path)
        self.iterator_class = Iterator.get_iterator_class(file_format)

    @property
    def data_files(self):
        paths = []

        for root, _, filenames in os.walk(self.directory_path):
            for filename in filenames:
                paths.append(os.path.join(root, filename))

        return [LocalDataFile(path, self.iterator_class) for path in paths]



from akagi.data_source import DataSource


class URLDataSource(DataSource):
    @classmethod
    def for_url(cls, url, file_format=FileFormat.CSV):
        bundle = URLDataFileBundle.for_url(url, file_format=file_format)

        return URLDataSource(bundle)




from akagi.data_file_bundle import DataFileBundle
from akagi.iterator import Iterator, FileFormat

from six.moves import urllib


class URLDataFileBundle(DataFileBundle):
    def __init__(self, urlstr, file_format=FileFormat.CSV):
        self.url = urllib.parse.urlparse(urlstr)
        self.file_format = file_format

    @property
    def data_files(self):
        if self.url.scheme in ['http', 'https']:
            return URLDataFile(url)
        # elif self.url.scheme in ['ftp', 'sftp']:
            # return



import os
import six

import MySQLdb

from akagi.data_source import DataSource
from akagi.query import Query
from akagi.log import logger


class MySQLDataSource(DataSource):
    '''MySQLDataSource replesents a set of row data as a result of query param.
    '''

    @classmethod
    def for_query(cls, query, db_conf={}):
        query = Query(query)

        return MySQLDataSource(query, db_conf)

    def __init__(self, query, db_conf={}):
        self.query = query
        self.__db_conf = db_conf
        self.__cursor = None

    @property
    def _connection(self):
        return MySQLdb.connect(**self._db_conf)

    def __iter__(self):
        self.__result = []
        logger.info("Query sent to Redshift")
        logger.info("\n" + self.query.body + "\n")

        self._cursor.execute(self.query.body)

        return iter(self._cursor.fetchall())

    @property
    def _cursor(self):
        if self.__cursor is None:
            self.__cursor = self._connection.cursor()

        return self.__cursor

    @property
    def _db_conf(self):
        conf = {
                'host':         self.__db_conf.get('host') or os.getenv('MYSQL_DB_HOST', 'localhost'),
                'user':         self.__db_conf.get('user') or os.getenv('MYSQL_DB_USER'),
                'passwd':       self.__db_conf.get('password') or os.getenv('MYSQL_DB_PASS'),
                'db':           self.__db_conf.get('db') or os.getenv('MYSQL_DB_PASS', os.getenv('USER')),
                'port':         self.__db_conf.get('port') or os.getenv('MYSQL_DB_PORT', 3306),
                'unix_socket':  self.__db_conf.get('unix_socket') or os.getenv('MYSQL_DB_SOCKET')
                }
        return {k: v for k, v in six.iteritems(conf) if v is not None}

    def __exit__(self, *exc):
        return False



# -*- coding: utf-8 -*-

from akagi.data_files import s3_data_file
from akagi.data_files import local_data_file

from akagi.data_files.s3_data_file import S3DataFile
from akagi.data_files.local_data_file import LocalDataFile



from __future__ import absolute_import

# -*- coding: utf-8 -*-
__author__ = """Yuichiro Someya"""
__email__ = 'ayemos.y@gmail.com'
__version__ = '0.1.15'


from akagi import data_file
from akagi import data_files

from akagi import data_file_bundle
from akagi import data_file_bundles

from akagi import iterators



from akagi.data_source import DataSource
from akagi.data_file_bundles import LocalDataFileBundle
from akagi.iterator import FileFormat


class LocalDataSource(DataSource):
    '''LocalDataSource replesents a set of files on local file system.
    '''

    @classmethod
    def for_path(cls, path, file_format=FileFormat.BINARY):
        bundle = LocalDataFileBundle(path, file_format)

        return LocalDataSource(bundle)

    def __init__(self, bundle):
        self.bundle = bundle

    def __exit__(self, *exc):
        return False



import six
from six import BytesIO

import gzip


def gzip_compress(data):
    if six.PY2:
        out_io = BytesIO()
        gzip.GzipFile(fileobj=out_io, mode='wb').write(data)
        out_io.seek(0)
        return out_io.read()
    else:
        return gzip.compress(data)


def gzip_decompress(data):
    if six.PY2:
        in_io = BytesIO()
        in_io.write(data)
        in_io.seek(0)
        return gzip.GzipFile(fileobj=in_io, mode='rb').read()
    else:
        return gzip.decompress(data)



import csv
from six import StringIO
import six


class CSVIterator(object):
    def __init__(self, content, skip_errors=True):
        self.content = content
        self._skip_erros = skip_errors
        self._iterator = None

    @classmethod
    def open_file(cls, path):
        return open(path, newline='')

    @classmethod
    def decode(cls, content):
        if six.PY2:
            return content
        else:
            return content.decode('utf-8')

    def __next__(self):
        try:
            n = self._iterator.__next__()
            return n
        except StopIteration as e:
            raise e
        except Exception as e:
            if self._skip_errors:
                return self._iterator.__next__()
            else:
                raise e

    def __iter__(self):
        self._iterator = csv.reader(StringIO(self.content), escapechar='\\')
        return self



# -*- coding: utf-8 -*-

from akagi.data_file_bundles import s3_data_file_bundle
from akagi.data_file_bundles import local_data_file_bundle

from akagi.data_file_bundles.s3_data_file_bundle import S3DataFileBundle
from akagi.data_file_bundles.local_data_file_bundle import LocalDataFileBundle



from itertools import chain


class DataFileBundle(object):
    '''DataFileBundle is an base class of all data file bundles
    '''

    def data_files(self):
        raise NotImplementedError

    def __iter__(self):
        return iter(chain(*self.data_files))

    def __enter__(self):
        return self

    def __exit__(self, *exc_type):
        self.clear()
        return False

    def clear(self):
        raise NotImplementedError



import re
import six


class Query(object):
    def __init__(self, body):
        self._body = body

    def wrap(self, query):
        raise NotImplementedError

    @property
    def body(self):
        if six.PY2:
            return self._body.decode('utf-8')
        else:
            return self._body

    def __str__(self):
        return self.body


class UnloadQuery(Query):
    def __init__(self, body, bundle, sort=False):
        super(UnloadQuery, self).__init__(body)

        self.bundle = bundle
        self.sort = sort

    @classmethod
    def wrap(cls, query_str, bundle, sort=False):
        return UnloadQuery(query_str, bundle, sort)

    @property
    def sql(self):
        return """
unload ('%(query)s')
to '%(bundle_url)s'
credentials '%(credential_string)s'
gzip
allowoverwrite
parallel %(enable_sort)s
delimiter ',' escape addquotes
            """ % ({
                'query': re.sub(r"'", "\\\\'", self.body),
                'bundle_url': self.bundle.url,
                'credential_string': self.bundle.credential_string,
                'enable_sort': "on" if self.sort else "off"
                })



import boto3

from six import BytesIO

from akagi.utils import gzip_decompress
from akagi.data_file import DataFile


class S3DataFile(DataFile):
    def __init__(self, obj, iterator_class):
        self.obj = obj
        self.iterator_class = iterator_class

        self.__s3 = None

    @property
    def filename(self):
        return self.key

    @property
    def key(self): return self.obj.key

    @property
    def raw_content(self):
        out_io = BytesIO()
        self._s3.Object(self.obj.bucket_name, self.obj.key).download_fileobj(out_io)
        out_io.seek(0)

        return out_io.read()

    @property
    def content(self):
        if self._is_gzip():
            return gzip_decompress(self.raw_content)
        else:
            return self.raw_content

    def __iter__(self):
        cls = self.iterator_class
        return iter(cls(cls.decode(self.content)))

    @property
    def _s3(self):
        if self.__s3 is None:
            self.__s3 = boto3.resource('s3')

        return self.__s3



import re
import boto3

from datetime import datetime

from akagi.iterator import Iterator, FileFormat
from akagi.data_files.s3_data_file import S3DataFile
from akagi.data_file_bundle import DataFileBundle
from akagi.log import logger


class S3DataFileBundle(DataFileBundle):
    @classmethod
    def for_table(self, bucket_name, schema, table, bucket_prefix='/'):
        timestamp = datetime.now().strftime("%Y%m%d_%H%M%S")
        prefix = "%(bucket_prefix)s/%(schema)s_export/%(table)s/%(timestamp)s/" % locals()

        return S3DataFileBundle(bucket_name, prefix=prefix, file_format=FileFormat.CSV)

    def __init__(self, bucket_name, prefix=None, keys=None, file_format=FileFormat.CSV):
        self.bucket_name = bucket_name
        self.prefix = self._normalize_path(prefix)
        self.keys = keys

        if self.prefix == None and self.keys == None:
            raise Exception("Either prefix or keys must be set.")

        self.file_format = file_format
        self.iterator_class = Iterator.get_iterator_class(file_format)

        self.__s3 = None

    @property
    def data_files(self):
        if self.prefix is not None:
            return [S3DataFile(obj, self.iterator_class)
                    for obj in self._bucket.objects.filter(Prefix=self.prefix)]
        else:
            return [S3DataFile(self._bucket.Object(key), self.iterator_class)
                    for key in self.keys]

    def clear(self):
        for obj in self._bucket.objects.filter(Prefix=self.prefix):
            logger.info("Deleting intermediate object on s3: %(key)s" % ({"key": obj.key}))
            obj.delete()

    @property
    def url(self):
        loc = "%(bucket_name)s/%(prefix)s" % ({
            'bucket_name': self.bucket_name,
            'prefix': self.prefix
            })
        loc = self._normalize_path(loc)

        return "s3://%(loc)s" % locals()

    @property
    def credential_string(self):
        credentials = []

        if self._credential is not None:
            if self._credential.access_key:
                credentials.append("aws_access_key_id=%s" % (self._credential.access_key))

            if self._credential.secret_key:
                credentials.append("aws_secret_access_key=%s" % (self._credential.secret_key))

            if self._credential.token:
                credentials.append("token=%s" % (self._credential.token))

        return ';'.join(credentials)

    @property
    def _credential(self):
        session = boto3.session.Session()
        return session.get_credentials()

    @property
    def _bucket(self):
        return self._s3.Bucket(self.bucket_name)

    @property
    def _s3(self):
        if self.__s3 is None:
            self.__s3 = boto3.resource('s3')

        return self.__s3

    def _normalize_path(self, path):
        if path is None:
            return None
        else:
            return re.sub(r'^/', '', re.sub(r'\/{2,}', '/', path))
